import inspect
from collections.abc import Callable, Iterable
from typing import Any

import pandas as pd

from pyranges.core.names import (
    BY_ENTRY_IN_KWARGS,
    PRESERVE_INDEX_COLUMN,
    RANGE_COLS,
    SKIP_IF_DF_EMPTY_DEFAULT,
    SKIP_IF_DF_EMPTY_TYPE,
    SKIP_IF_EMPTY_ANY,
    SKIP_IF_EMPTY_BOTH,
    SKIP_IF_EMPTY_LEFT,
    SKIP_IF_EMPTY_RIGHT,
    VALID_BY_TYPES,
    VALID_OVERLAP_TYPE,
    BinaryOperation,
    UnaryOperation,
)
from pyranges.core.pyranges_helpers import arg_to_list
from pyranges.core.tostring import tostring
from pyranges.range_frame.range_frame_validator import InvalidRangesReason


def should_skip_operation(df: pd.DataFrame, *, df2: pd.DataFrame, skip_if_empty: SKIP_IF_DF_EMPTY_TYPE) -> bool:
    """Whether to skip operation because one or more dfs are empty."""
    if df.empty and df2.empty:
        return skip_if_empty in {SKIP_IF_EMPTY_BOTH, SKIP_IF_EMPTY_ANY, SKIP_IF_EMPTY_LEFT, SKIP_IF_EMPTY_RIGHT}
    if df.empty:
        return skip_if_empty in {SKIP_IF_EMPTY_LEFT, SKIP_IF_EMPTY_ANY}
    if df2.empty:
        return skip_if_empty in {SKIP_IF_EMPTY_RIGHT, SKIP_IF_EMPTY_ANY}
    return False


class RangeFrame(pd.DataFrame):
    """Class for range based operations.

    A table with Start and End columns. Parent class of PyRanges. Subclass of pandas DataFrame.
    """

    def __new__(cls, *args, **kwargs) -> "RangeFrame | pd.DataFrame":  # type: ignore[misc]
        """Create a new instance of a PyRanges object."""
        # __new__ is a special static method used for creating and
        # returning a new instance of a class. It is called before
        # __init__ and is typically used in scenarios requiring
        # control over the creation of new instances

        if not args:
            return super().__new__(cls)
        if not kwargs:
            return super().__new__(cls)

        df = pd.DataFrame(kwargs.get("data") or (args[0]))

        missing_any_required_columns = not set(RANGE_COLS).issubset(df.columns)
        if missing_any_required_columns:
            return df

        return super().__new__(cls)

    @property
    def _constructor(self) -> type:
        return RangeFrame

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        missing_columns = [c for c in RANGE_COLS if c not in self.columns]
        if missing_columns:
            msg = f"Missing required columns: {missing_columns}"
            raise ValueError(msg)

    def __str__(
        self,
        **kwargs: int | None,
    ) -> str:  # , max_col_width: int | None = None, max_total_width: int | None = None) -> str:
        """Return string representation."""
        str_repr = tostring(
            self,
            max_col_width=kwargs.get("max_col_width"),
            max_total_width=kwargs.get("max_total_width"),
        )
        if reasons := InvalidRangesReason.formatted_reasons_list(self):
            return f"{str_repr}\nInvalid ranges:\n{reasons}"
        return str_repr

    def __repr__(self, max_col_width: int | None = None, max_total_width: int | None = None) -> str:
        return self.__str__(max_col_width=max_col_width, max_total_width=max_total_width)

    def overlap(
        self,
        other: "RangeFrame",
        how: VALID_OVERLAP_TYPE = "first",
        by: VALID_BY_TYPES = None,
    ) -> "RangeFrame":
        """Find intervals in self overlapping other..

        Parameters
        ----------
        other
            Other ranges to find overlaps with.
        how
            How to find overlaps. "first" finds the first overlap, "containment" finds all overlaps
            where self is contained in other, and "all" finds all overlaps.
        by:
            Grouping columns. If None, all columns are used.

        Returns
        -------
        RangeFrame
            RangeFrame with overlapping ranges.

        Examples
        --------
        >>> import pyranges as pr
        >>> r = pr.RangeFrame({"Start": [1, 1, 2, 2], "End": [3, 3, 5, 4], "Id": list("abad")})
        >>> r
          index  |      Start      End  Id
          int64  |      int64    int64  object
        -------  ---  -------  -------  --------
              0  |          1        3  a
              1  |          1        3  b
              2  |          2        5  a
              3  |          2        4  d
        RangeFrame with 4 rows, 3 columns, and 1 index columns.

        >>> r2 = pr.RangeFrame({"Start": [0, 2], "End": [1, 20], "Id": list("ad")})
        >>> r2
          index  |      Start      End  Id
          int64  |      int64    int64  object
        -------  ---  -------  -------  --------
              0  |          0        1  a
              1  |          2       20  d
        RangeFrame with 2 rows, 3 columns, and 1 index columns.

        >>> r.overlap(r2, how="first")
          index  |      Start      End  Id
          int64  |      int64    int64  object
        -------  ---  -------  -------  --------
              0  |          1        3  a
              1  |          1        3  b
              2  |          2        5  a
              3  |          2        4  d
        RangeFrame with 4 rows, 3 columns, and 1 index columns.

        >>> r.overlap(r2, how="containment")
          index  |      Start      End  Id
          int64  |      int64    int64  object
        -------  ---  -------  -------  --------
              2  |          2        5  a
              3  |          2        4  d
        RangeFrame with 2 rows, 3 columns, and 1 index columns.

        >>> r.overlap(r2, how="all")
          index  |      Start      End  Id
          int64  |      int64    int64  object
        -------  ---  -------  -------  --------
              0  |          1        3  a
              1  |          1        3  b
              2  |          2        5  a
              3  |          2        4  d
        RangeFrame with 4 rows, 3 columns, and 1 index columns.

        >>> r.overlap(r2, how="all", by="Id")
          index  |      Start      End  Id
          int64  |      int64    int64  object
        -------  ---  -------  -------  --------
              3  |          2        4  d
        RangeFrame with 1 rows, 3 columns, and 1 index columns.

        """
        from pyranges.methods.overlap import _overlap

        return self.apply_pair(other, _overlap, how=how, by=by)

    def apply_single(
        self,
        function: UnaryOperation,
        by: VALID_BY_TYPES,
        *,
        preserve_index: bool = False,
        **kwargs: Any,
    ) -> "RangeFrame":
        """Call a function on a RangeFrame.

        Parameters
        ----------
        function: Callable
            Function to call.

        by: str or list of str
            Group by these columns.

        preserve_index: bool
            Preserve the original index. Only valid if the function returns the index cols.

        kwargs: dict
            Passed to function.

        """
        assert_valid_ranges(function, self)

        if not by:
            return _mypy_ensure_rangeframe(function(self, **kwargs))
        by = arg_to_list(by)

        f = _with_group_keys_to_kwargs(by)(function)
        if not preserve_index:
            return _mypy_ensure_rangeframe(self.groupby(by).apply(f, by=by, **kwargs).reset_index(drop=True))

        result = (
            self.assign(**{PRESERVE_INDEX_COLUMN: lambda df: df.index})
            .groupby(by)
            .apply(
                f,
                by=by,
                **kwargs,
            )
            .reset_index(drop=True)
        )
        result = result.set_index(PRESERVE_INDEX_COLUMN)
        if isinstance(self.index, pd.MultiIndex):
            result.index.names = self.index.names
        else:
            result.index.name = self.index.name
        return _mypy_ensure_rangeframe(result)

    def apply_pair(
        self,
        other: "RangeFrame",
        function: BinaryOperation,
        by: VALID_BY_TYPES = None,
        skip_if_empty: SKIP_IF_DF_EMPTY_TYPE = SKIP_IF_DF_EMPTY_DEFAULT,
        **kwargs,
    ) -> "RangeFrame":
        """Call a function on two RangeFrames.

        Parameters
        ----------
        other: RangeFrame
            Other RangeFrame.

        function: Callable
            Function to call.

        by: str or list of str, default None
            Group by these columns.

        kwargs: dict
            Passed to function.

        skip_if_empty:
            Whether to skip the operations if one of the dataframes is empty for a particular group.

        Examples
        --------
        >>> import pyranges as pr
        >>> r = pr.RangeFrame({"Start": [1, 1, 4, 2], "End": [3, 3, 5, 4], "Id": list("abad")})
        >>> bad, ok = r, r.copy()
        >>> bad.loc[0, "Start"] = -1  # make r invalid
        >>> bad.apply_pair(ok, lambda x, y: x)
        Traceback (most recent call last):
        ...
        ValueError: Cannot perform function on invalid ranges (function was bad.apply_pair(ok, lambda x, y: x)).
        >>> from pyranges.methods.overlap import _overlap
        >>> ok.apply_pair(bad, _overlap)
        Traceback (most recent call last):
        ...
        ValueError: Cannot perform function on invalid ranges (function was _overlap).

        """
        assert_valid_ranges(function, self, other)
        if by is None:
            return _mypy_ensure_rangeframe(function(self, df2=other, **kwargs))

        by = arg_to_list(by)
        results = []
        empty = RangeFrame(columns=other.columns)
        others = dict(list(other.groupby(by)))

        for key, _df in self.groupby(by):
            odf = others.get(key, empty)

            if should_skip_operation(_df, df2=odf, skip_if_empty=skip_if_empty):
                continue
            results.append(
                function(
                    _mypy_ensure_rangeframe(_df),
                    df2=_mypy_ensure_rangeframe(odf),
                    **(
                        kwargs
                        | {BY_ENTRY_IN_KWARGS: dict(zip(by, [key] if not isinstance(key, tuple) else key, strict=True))}
                    ),
                ),
            )

        if not results:
            return _mypy_ensure_rangeframe(RangeFrame(columns=self.columns))
        return _mypy_ensure_rangeframe(pd.concat(results))

    def sort_by_position(self) -> "RangeFrame":
        """Sort by Start and End columns."""
        return _mypy_ensure_rangeframe(self.sort_values(RANGE_COLS))

    def reasons_why_frame_is_invalid(self) -> list[InvalidRangesReason] | None:  # noqa: D102
        __doc__ = InvalidRangesReason.is_invalid_ranges_reasons.__doc__  # noqa: A001, F841

        return InvalidRangesReason.is_invalid_ranges_reasons(self)

    def copy(self, *args, **kwargs) -> "RangeFrame":  # noqa: D102
        return _mypy_ensure_rangeframe(super().copy(*args, **kwargs))

    def drop(self, *args, **kwargs) -> "RangeFrame | None":  # type: ignore[override]  # noqa: D102
        return self.__class__(super().drop(*args, **kwargs))

    def drop_and_return[T: "RangeFrame"](self: T, *args: Any, **kwargs: Any) -> T:  # noqa: PYI019, D102
        kwargs["inplace"] = False
        return self.__class__(super().drop(*args, **kwargs))

    def reindex(self, *args, **kwargs) -> "RangeFrame":  # noqa: D102
        return self.__class__(super().reindex(*args, **kwargs))


def _mypy_ensure_rangeframe(r: pd.DataFrame) -> "RangeFrame":
    result = RangeFrame(r)
    if not isinstance(result, RangeFrame):
        msg = f"Expected RangeFrame, got {type(result)}"
        raise TypeError(msg)
    return result


def assert_valid_ranges(function: Callable, *args: "RangeFrame") -> None:
    """Raise ValueError because function is called on invalid ranges."""
    if any(r.reasons_why_frame_is_invalid() for r in args):
        is_not_lambda = function.__name__ != "<lambda>"
        function_repr = function.__name__ if is_not_lambda else inspect.getsource(function).strip()
        msg = f"Cannot perform function on invalid ranges (function was {function_repr})."
        raise ValueError(msg)


def _with_group_keys_to_kwargs(by: list[str]) -> Callable:
    def decorator(func) -> Callable:
        def wrapper(group, **kwargs) -> Any:
            # Extract the group keys and add them to kwargs
            names = group.name if isinstance(group.name, Iterable) and not isinstance(group.name, str) else [group.name]
            kwargs[BY_ENTRY_IN_KWARGS] = dict(zip(by, names, strict=True))
            return func(group, **kwargs)

        return wrapper

    return decorator



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from pyranges.core.names import (
    BY_ENTRY_IN_KWARGS,
    CHROM_COL,
    END_COL,
    FORWARD_STRAND,
    REVERSE_STRAND,
    START_COL,
    STRAND_COL,
    TEMP_END_COL,
    TEMP_INDEX_COL,
    TEMP_MAX_COL,
    TEMP_MIN_COL,
    TEMP_START_COL,
)

if TYPE_CHECKING:
    from pyranges import PyRanges


def _subseq(
    df: "PyRanges",
    *,
    by: list[str],
    start: int = 0,
    end: int | None = None,
    **kwargs,
) -> pd.DataFrame:
    if df.empty:
        return df

    df = df.copy()
    orig_order = df.index.copy()
    df.insert(0, TEMP_INDEX_COL, orig_order)

    by = [col for col in by if col not in [CHROM_COL, STRAND_COL]]
    by_argument_given = bool(by)
    by = by or [TEMP_INDEX_COL]

    strand = kwargs.get(BY_ENTRY_IN_KWARGS, {}).get("Strand")

    # at this point, strand is False if 1. spliced_subsequence was called with use_strand=False or
    #                                   2. it was called with use_strand='auto' and not self.valid_strand
    # or it can be '+' or '-' if:
    #  1. it was input as True to spliced_subsequence and passed  to pyrange_apply_single as True,
    #     which updates it to '-' or '+' before calling _spliced_subseq, or
    #  2. it was called with strand=None and self is stranded

    strand = strand if strand else FORWARD_STRAND

    # creating j which holds the boundaries per group
    # j contains one row per group; columns: Start  End (+ by columns); indexed by __i__
    agg_dict = {TEMP_INDEX_COL: "first", START_COL: "min", END_COL: "max"} | {k: "first" for k in by}

    if by_argument_given:
        j = (
            df.groupby(by, dropna=False)[[START_COL, END_COL, TEMP_INDEX_COL, *by]]
            .agg(agg_dict)
            .set_index(TEMP_INDEX_COL)
        )
    else:
        j = df.groupby(by, dropna=False)[[START_COL, END_COL, TEMP_INDEX_COL]].agg(agg_dict).set_index(TEMP_INDEX_COL)
        j.insert(0, TEMP_INDEX_COL, j.index)
        j.index.name = None

    _end = end if end is not None else (j.End - j.Start).max()

    # below we add columns starts, ends to j
    # start and ends define the desired start and end, with one row per group (transcript).
    # they may be out of bounds of the interval, though (this is dealt with later)

    if (strand == FORWARD_STRAND and start >= 0) or (strand == REVERSE_STRAND and start < 0):
        j.loc[:, TEMP_START_COL] = j.Start + abs(start)
    else:
        j.loc[:, TEMP_START_COL] = j.End - abs(start)

    if (strand == FORWARD_STRAND and _end >= 0) or (strand == REVERSE_STRAND and _end < 0):
        j.loc[:, TEMP_END_COL] = j.Start + abs(_end)
    else:
        j.loc[:, TEMP_END_COL] = j.End - abs(_end)

    if strand == REVERSE_STRAND:
        j = j.rename(columns={TEMP_END_COL: TEMP_MIN_COL, TEMP_START_COL: TEMP_MAX_COL})
    else:
        j = j.rename(columns={TEMP_START_COL: TEMP_MIN_COL, TEMP_END_COL: TEMP_MAX_COL})

    # I'm maintaing the original row order
    _df = df.merge(j[[*by, TEMP_MIN_COL, TEMP_MAX_COL]], on=by).set_index(TEMP_INDEX_COL).loc[orig_order]

    # instead of simply using starts and ends as computed above, we're dealing here with potential out of bounds:
    r = _df[~((_df[START_COL] >= _df[TEMP_MAX_COL]) | (_df[END_COL] <= _df[TEMP_MIN_COL]))].copy()
    r.loc[:, START_COL] = np.maximum(r.Start, r[TEMP_MIN_COL])
    r.loc[:, END_COL] = np.minimum(r.End, r[TEMP_MAX_COL])

    return r.drop([TEMP_MIN_COL, TEMP_MAX_COL], axis=1)



from typing import TYPE_CHECKING

import pandas as pd
from ncls import NCLS  # type: ignore[import]
from sorted_nearest import (  # type: ignore[import]
    nearest_next_nonoverlapping,
    nearest_nonoverlapping,
    nearest_previous_nonoverlapping,
)

import pyranges.core.empty
from pyranges import PyRanges
from pyranges.core.names import END_COL, START_COL
from pyranges.methods.sort import sort_one_by_one

if TYPE_CHECKING:
    from numpy.typing import ArrayLike, NDArray
    from pandas import DataFrame


def _insert_distance(df2: pd.DataFrame, dist: "NDArray | int", suffix: str) -> pd.DataFrame:
    if "Distance" not in df2:
        distance_column_name = "Distance"
    elif "Distance" + suffix not in df2:
        distance_column_name = "Distance" + suffix
    else:
        i = 1
        while "Distance" + str(i) in df2:
            i += 1
        distance_column_name = "Distance" + str(i)

    df2.insert(
        df2.shape[1],
        distance_column_name,
        pd.Series(dist, index=df2.index).fillna(-1).astype(int),
    )

    return df2


def _overlapping_for_nearest(df: pd.DataFrame, df2: pd.DataFrame, suffix: str) -> tuple[pd.DataFrame, pd.DataFrame]:
    nearest_df = pd.DataFrame(columns="Chromosome Start End Strand".split())

    it = NCLS(df2.Start.to_numpy(), df2.End.to_numpy(), df2.index.to_numpy())

    idx_self, idx_other = it.first_overlap_both(
        df[START_COL].to_numpy(),
        df[END_COL].to_numpy(),
        df.index.to_numpy(),
    )
    self, other = df.reindex(idx_self), df2.reindex(idx_other)

    if not other.empty:
        idxs = self.index
        original_idx = df.index.copy(deep=True)
        missing_idxs = ~original_idx.isin(idxs)
        missing_overlap = df.index[missing_idxs]

        df_to_find_nearest_in = df.reindex(missing_overlap)

        odf = df2.reindex(other.index)
        odf.index = idxs
        sdf = df.reindex(idxs)

        nearest_df = sdf.join(odf, rsuffix=suffix)
        nearest_df = _insert_distance(nearest_df, 0, suffix)
    else:
        df_to_find_nearest_in = df

    return nearest_df, df_to_find_nearest_in


def _next_nonoverlapping(
    left_ends: pd.Series,
    right_starts: pd.Series,
    right_indexes: "ArrayLike",
) -> tuple["NDArray", "NDArray"]:
    left_ends = left_ends.sort_values()
    right_starts = right_starts.sort_values()
    r_idx, dist = nearest_next_nonoverlapping(left_ends.to_numpy() - 1, right_starts.to_numpy(), right_indexes)
    r_idx = pd.Series(r_idx, index=left_ends.index).sort_index().to_numpy()
    dist = pd.Series(dist, index=left_ends.index).sort_index().to_numpy()

    return r_idx, dist


def _previous_nonoverlapping(left_starts: pd.Series, right_ends: pd.Series) -> tuple["NDArray", "NDArray"]:
    left_starts = left_starts.sort_values()
    right_ends = right_ends.sort_values()
    r_idx, dist = nearest_previous_nonoverlapping(
        left_starts.to_numpy(),
        right_ends.to_numpy() - 1,
        right_ends.index.to_numpy(),
    )

    r_idx = pd.Series(r_idx, index=left_starts.index).sort_index().to_numpy()
    dist = pd.Series(dist, index=left_starts.index).sort_index().to_numpy()

    return r_idx, dist


def _nearest(df: "DataFrame", df2: "DataFrame", **kwargs) -> pd.DataFrame:
    if df.empty or df2.empty:
        return PyRanges()

    overlap = kwargs["overlap"]
    how = kwargs["how"]
    suffix = kwargs["suffix"]

    if how == "upstream":
        strand = df.Strand.iloc[0]
        how = {"+": "previous", "-": "next"}[strand]
    elif how == "downstream":
        strand = df.Strand.iloc[0]
        how = {"+": "next", "-": "previous"}[strand]

    df2 = df2.reset_index(drop=True)

    if overlap:
        nearest_df, df_to_find_nearest_in = _overlapping_for_nearest(df, df2, suffix)
    else:
        df_to_find_nearest_in = df
        nearest_df = pyranges.core.empty.empty()

    df = pyranges.core.empty.empty_df()
    if not df_to_find_nearest_in.empty:
        df_to_find_nearest_in = sort_one_by_one(df_to_find_nearest_in, "Start", "End")
        df2 = sort_one_by_one(df2, "Start", "End")
        df_to_find_nearest_in.index = pd.Index(range(len(df_to_find_nearest_in)))

        if how == "next":
            r_idx, dist = _next_nonoverlapping(df_to_find_nearest_in.End, df2.Start, df2.index.to_numpy())
        elif how == "previous":
            r_idx, dist = _previous_nonoverlapping(df_to_find_nearest_in.Start, df2.End)
        else:
            previous_r_idx, previous_dist = _previous_nonoverlapping(df_to_find_nearest_in.Start, df2.End)

            next_r_idx, next_dist = _next_nonoverlapping(df_to_find_nearest_in.End, df2.Start, df2.index.to_numpy())

            r_idx, dist = nearest_nonoverlapping(previous_r_idx, previous_dist, next_r_idx, next_dist)

        df2 = df2.reindex(r_idx)

        df2.index = df_to_find_nearest_in.index

        df2 = _insert_distance(df2, dist, suffix)

        _r_idx = pd.Series(r_idx, index=df2.index)
        df_to_find_nearest_in = df_to_find_nearest_in.drop(_r_idx[_r_idx == -1].index)

        df = df_to_find_nearest_in.join(df2, rsuffix=suffix)

    if overlap and "df" in locals() and not df.empty and not nearest_df.empty:
        df = pd.concat([nearest_df, df], sort=False)
    elif overlap and not nearest_df.empty:
        df = nearest_df

    return PyRanges(df.drop("Chromosome" + suffix, axis=1))



import pandas as pd
import pyranges as pr


def test_spliced_subsequence_use_strand_false():
    p = pr.PyRanges(
        {
            "Chromosome": [1, 1, 2, 2, 3],
            "Strand": ["+", "+", "-", "-", "+"],
            "Start": [1, 40, 10, 70, 140],
            "End": [11, 60, 25, 80, 152],
            "transcript_id": ["t1", "t1", "t2", "t2", "t3"],
        },
    )

    result = p.spliced_subsequence(0, 5, use_strand=False, transcript_id="transcript_id")

    _expected_result = {
        "Chromosome":{
            0:1,
            2:2,
            4:3
        },
        "Strand":{
            0:"+",
            2:"-",
            4:"+"
        },
        "Start":{
            0:1,
            2:10,
            4:140,
        },
        "End":{
            0:6,
            2:15,
            4:145
        },
        "transcript_id":{
            0:"t1",
            2:"t2",
            4:"t3"
        }
    }
    expected_result = pd.DataFrame(_expected_result)
    print(result)
    print(expected_result)
    pd.testing.assert_frame_equal(result, expected_result)



def test_spliced_subsequence_without_transcript_id():
    p = pr.PyRanges(
        {
            "Chromosome": [1, 1, 2, 2, 3],
            "Strand": ["+", "+", "-", "-", "+"],
            "Start": [1, 40, 10, 70, 140],
            "End": [11, 60, 25, 80, 152],
            "transcript_id": ["t1", "t1", "t2", "t2", "t3"],
        },
    )

    result = p.spliced_subsequence(0, 5, use_strand=False)

    _expected_result = {
        "Chromosome":{
            0:1,
            1:1,
            2:2,
            3:2,
            4:3
        },
        "Strand":{
            0:"+",
            1:"+",
            2:"-",
            3:"-",
            4:"+"
        },
        "Start":{
            0:1,
            1:40,
            2:10,
            3:70,
            4:140
        },
        "End":{
            0:6,
            1:45,
            2:15,
            3:75,
            4:145
        },
        "transcript_id":{
            0:"t1",
            1:"t1",
            2:"t2",
            3:"t2",
            4:"t3"
        }
    }


    expected_result = pd.DataFrame(_expected_result)
    print(result)
    print(expected_result)
    pd.testing.assert_frame_equal(result, expected_result)




from typing import TYPE_CHECKING

import pandas as pd
from ncls import NCLS  # type: ignore[import]

from pyranges.core.names import TEMP_NUM_COL

if TYPE_CHECKING:
    from pyranges import RangeFrame


def _subtraction(df: "RangeFrame", df2: "RangeFrame", **_) -> pd.DataFrame:
    if df2.empty or df.empty:
        return df

    o = NCLS(df2.Start.to_numpy(), df2.End.to_numpy(), df2.index.to_numpy())

    idx_self, new_starts, new_ends = o.set_difference_helper(
        df.Start.values,
        df.End.values,
        df.index.values,
        df[TEMP_NUM_COL].to_numpy(),
    )

    missing_idx = pd.Index(df.index).difference(idx_self)

    idx_to_drop = new_starts != -1

    new_starts = new_starts[idx_to_drop]
    new_ends = new_ends[idx_to_drop]

    idx_self = idx_self[idx_to_drop]
    new_starts = pd.Series(new_starts, index=idx_self)
    new_ends = pd.Series(new_ends, index=idx_self)

    _df = df.reindex(missing_idx.union(idx_self)).sort_index()
    new_starts = new_starts.sort_index()
    new_ends = new_ends.sort_index()

    if len(idx_self):
        _df.loc[_df.index.isin(idx_self), "Start"] = new_starts.to_numpy()
        _df.loc[_df.index.isin(idx_self), "End"] = new_ends.to_numpy()

    return _df



import inspect
import warnings
from typing import TYPE_CHECKING

from pyranges.core.names import (
    CHROM_COL,
    REVERSE_STRAND,
    STRAND_BEHAVIOR_AUTO,
    STRAND_BEHAVIOR_IGNORE,
    STRAND_BEHAVIOR_OPPOSITE,
    STRAND_BEHAVIOR_SAME,
    STRAND_COL,
    STRICT_STRAND_BEHAVIOR_TYPE,
    TEMP_STRAND_COL,
    USE_STRAND_AUTO,
    VALID_BY_OPTIONS,
    VALID_STRAND_BEHAVIOR_OPTIONS,
    VALID_STRAND_BEHAVIOR_TYPE,
    VALID_USE_STRAND_OPTIONS,
    VALID_USE_STRAND_TYPE,
)

if TYPE_CHECKING:
    import pandas as pd

    from pyranges import PyRanges


def validate_and_convert_use_strand(self: "PyRanges", use_strand: VALID_USE_STRAND_TYPE) -> bool:
    """Validate and convert strand option.

    Used inside various methods accepting a PyRanges as input. It processes use_strand option provided by the user upon
    calling the function. This function validates the option and converts it to a boolean value.

    If use_strand is True, an error will be raised if the Strand column is missing or invalid.
    If use_strand is 'auto', it will be converted to True if the Strand column is valid, otherwise False.
    Note that if use_strand is 'auto' and the Strand column is invalid, a warning will be shown.

    Parameters
    ----------
    self : PyRanges
        PyRanges object to be validated.

    use_strand : {'auto', True, False}
        The use_strand option provided by the user.


    Returns
    -------
    bool
        The converted boolean value of use_strand.

    """
    if use_strand not in VALID_USE_STRAND_OPTIONS:
        msg = f"Invalid use_strand option: {use_strand}"
        raise ValueError(msg)
    if use_strand is True and not self.has_strand:
        msg = "Cannot have use_strand=True when the Strand column is missing."
        raise ValueError(msg)
    if use_strand == USE_STRAND_AUTO:
        use_strand = self.strand_valid

        # If Strand column is invalid, set use_strand to False
        # but if there are intervals on the negative strand, warn user
        if not use_strand and contains_negative_strand(self):
            fn_name = stack[2].function if len(stack := inspect.stack()) > 2 else "unknown_function"  # noqa: PLR2004
            fn_name = stack[1].function if fn_name == "<module>" else fn_name
            msg = f"{fn_name}: '{USE_STRAND_AUTO}' use_strand treated as False due to invalid Strand values. Please use use_strand=False"
            warnings.warn(msg, stacklevel=4)
    return use_strand


def validate_and_convert_strand_behavior(
    self: "PyRanges",
    other: "PyRanges",
    strand_behavior: VALID_STRAND_BEHAVIOR_TYPE,
) -> STRICT_STRAND_BEHAVIOR_TYPE:
    """Validate and convert strand behavior.

    Used inside various methods accepting two PyRanges as input. It processes strand_behavior option provided
    by the user upon calling the function. This function validates the option and converts it to a valid unambiguous
    value.
    """
    if strand_behavior not in VALID_STRAND_BEHAVIOR_OPTIONS:
        msg = f"{VALID_STRAND_BEHAVIOR_OPTIONS} are the only valid values for strand_behavior. Was: {strand_behavior}"
        raise ValueError(msg)

    if not (self.strand_valid and other.strand_valid):
        if strand_behavior == STRAND_BEHAVIOR_AUTO:
            # If any Strand column is invalid, set strand_behavior to STRAND_BEHAVIOR_IGNORE
            # but warn user if both objects are Stranded and there are intervals on the negative strand
            strand_behavior = STRAND_BEHAVIOR_IGNORE
            if (
                self.has_strand
                and other.has_strand
                and (contains_negative_strand(self) or contains_negative_strand(other))
            ):
                fn_name = stack[2].function if len(stack := inspect.stack()) > 2 else "unknown_function"  # noqa: PLR2004
                fn_name = stack[1].function if fn_name == "<module>" else fn_name
                msg = (
                    f"{fn_name}: '{STRAND_BEHAVIOR_AUTO}' strand_behavior treated as {STRAND_BEHAVIOR_IGNORE} due to invalid Strand values. "
                    f"Please use strand_behavior={STRAND_BEHAVIOR_IGNORE}"
                )
                warnings.warn(msg, stacklevel=4)

        elif strand_behavior in (STRAND_BEHAVIOR_SAME, STRAND_BEHAVIOR_OPPOSITE):
            msg = f"Can only do {strand_behavior} strand operations when both PyRanges contain valid strand info."
            raise ValueError(msg)
    elif strand_behavior == STRAND_BEHAVIOR_AUTO:
        strand_behavior = STRAND_BEHAVIOR_SAME

    return strand_behavior


def group_keys_from_validated_strand_behavior(
    strand_behavior: VALID_STRAND_BEHAVIOR_TYPE,
    by: VALID_BY_OPTIONS = None,
) -> list[str]:
    """Return group keys based on strand behavior.

    If strand_behavior is 'same', return [CHROM_COL, STRAND_COL].
    If strand_behavior is 'opposite', return [CHROM_COL, TEMP_STRAND_COL].
    If strand_behavior is 'ignore', return [CHROM_COL].

    In each case, if a by argument is provided, it is appended/extended to the list of group keys.

    """
    strand_cols = (
        [STRAND_COL]
        if strand_behavior == STRAND_BEHAVIOR_SAME
        else ([TEMP_STRAND_COL] if strand_behavior == STRAND_BEHAVIOR_OPPOSITE else [])
    )

    return [CHROM_COL, *strand_cols, *arg_to_list(by)]


def strand_behavior_from_validated_use_strand(
    df: "PyRanges",
    use_strand: VALID_USE_STRAND_TYPE,
) -> VALID_STRAND_BEHAVIOR_TYPE:
    """Return strand_behavior based on use_strand bool.

    If strand is True, returns 'same' strand behavior, otherwise 'ignore' strand behavior.

    This function must be used with use_strand validated and converted to a boolean value.
    """
    return STRAND_BEHAVIOR_SAME if validate_and_convert_use_strand(df, use_strand) else STRAND_BEHAVIOR_IGNORE


def use_strand_from_validated_strand_behavior(
    self: "PyRanges",
    other: "PyRanges",
    strand_behavior: VALID_STRAND_BEHAVIOR_TYPE,
) -> bool:
    """Return use_strand based on strand_behavior.

    If strand_behavior is 'ignore', returns False, otherwise True if both PyRanges contain valid strand info.

    This function must be used with strand_behavior validated and converted to either 'same', 'opposite' or 'ignore'.
    """
    strand_behavior = validate_and_convert_strand_behavior(self, other, strand_behavior)
    return strand_behavior != STRAND_BEHAVIOR_IGNORE


def mypy_ensure_pyranges(df: "pd.DataFrame") -> "PyRanges":
    """Ensure df is a PyRanges.

    Helps mypy.
    """
    from pyranges import PyRanges

    if not isinstance(ret := PyRanges(df), PyRanges):
        msg = "Not a PyRanges"
        raise TypeError(msg)
    return ret


def arg_to_list(by: VALID_BY_OPTIONS) -> list[str]:
    """Convert a by or column_key argument to a list.

    If by is a string, it will be converted to a list.
    If by is None, an empty list will be returned.
    If by is already a list, it will be returned as is.
    """
    return [by] if isinstance(by, str) else ([*by] if by is not None else [])


def contains_negative_strand(self: "PyRanges") -> bool:
    """Return True if negative strand is present."""
    return self.has_strand and REVERSE_STRAND in self[STRAND_COL].to_numpy()
